import numpy as np

PI = np.pi
"""Pi constant."""

FREE_SPACE_PERMITTIVITY: float = 8.85418782e-12
"""Permittivity of free space [m-3 kg-1 s4 A2]."""

FREE_SPACE_PERMEABILITY: float = 4 * PI * 1e-7
"""Permeability of free space [m kg s-2 A-2]."""

FREE_SPACE_LIGHT_SPEED: float = 299792458.0
"""Speed of light in free space [m/s]."""



import numpy as np
from typing import Callable
import numpy.typing as npt
from scipy.special import spherical_jn as sjn, spherical_yn as syn
from scatsol.material import Material, Medium
import scatsol.utils
from scatsol.utils import ric_jnp, ric_jn, ric_h2n, ric_h2np


def mie_incident_field(
    xyz: npt.NDArray[np.float64],
    frequency: float,
    background: Material,
) -> tuple[npt.NDArray[np.complex128], npt.NDArray[np.complex128]]:
    Ertp = np.zeros_like(xyz, dtype=np.complex128)
    Hrtp = np.zeros_like(xyz, dtype=np.complex128)
    r, theta, phi = scatsol.utils.cart2spherical(xyz).T
    bg = Medium(background, frequency)
    t_cos, t_sin = np.cos(theta), np.sin(theta)
    p_cos, p_sin = np.cos(phi), np.sin(phi)
    exp = np.exp(-1j * bg.k * r * np.cos(theta))
    Ertp[:, 0] = t_sin * p_cos * exp
    Ertp[:, 1] = t_cos * p_cos * exp
    Ertp[:, 2] = -p_sin * exp
    Hrtp[:, 0] = t_sin * p_sin * exp / bg.eta
    Hrtp[:, 1] = t_cos * p_sin * exp / bg.eta
    Hrtp[:, 2] = p_cos * exp / bg.eta
    return Ertp, Hrtp


def mie_total_field(
    xyz: npt.NDArray[np.float64],
    radius: float,
    frequency: float,
    background: Material,
    sphere: Material | None = None,
    *,
    n: int = 50,
) -> tuple[npt.NDArray[np.complex128], npt.NDArray[np.complex128]]:
    Ertp = np.zeros_like(xyz, dtype=np.complex128)
    Hrtp = np.zeros_like(xyz, dtype=np.complex128)

    mask = (xyz**2).sum(axis=1) > radius**2

    bg = Medium(background, frequency)
    if sphere == None:
        an, bn = _an_bn_cond(bg.k, radius, n)
        Ertp[mask], Hrtp[mask] = _calculate_field(xyz[mask], bg.k, bg.eta, an, bn, ric_h2n, ric_h2np)
    else:
        s = Medium(sphere, frequency)
        an, bn, cn, dn = _an_bn_cn_dn_diel(bg.k, sphere.epsilon_r, sphere.mu_r, radius, n)
        Ertp[mask], Hrtp[mask] = _calculate_field(xyz[mask], bg.k, bg.eta, an, bn, ric_h2n, ric_h2np)
        Ertp[~mask], Hrtp[~mask] = _calculate_field(xyz[~mask], s.k, s.eta, cn, dn, ric_jn, ric_jnp)
    Ei, Hi = mie_incident_field(xyz[mask], frequency, background)
    Ertp[mask] += Ei
    Hrtp[mask] += Hi
    return Ertp, Hrtp


def _an_bn_cond(k: float, a: float, n: int) -> tuple[npt.NDArray[np.complex128], npt.NDArray[np.complex128]]:
    nn = np.arange(1, n + 1)
    scale_term = (2 * nn + 1) / (nn * (nn + 1))
    sj = sjn(nn, k * a)
    sy = syn(nn, k * a)
    sjp = sjn(nn, k * a, derivative=True)
    syp = syn(nn, k * a, derivative=True)
    bn = -(1j**-nn) * scale_term * sj / (sj - 1j * sy)
    an_num = sj + k * a * sjp
    an_den = an_num - 1j * (sy + k * a * syp)
    an = -(1j**-nn) * scale_term * an_num / an_den
    return an, bn


def _an_bn_cn_dn_diel(
    k_0: float, eps_r: float, mu_r: float, a: float, n: int
) -> tuple[
    npt.NDArray[np.complex128],
    npt.NDArray[np.complex128],
    npt.NDArray[np.complex128],
    npt.NDArray[np.complex128],
]:
    # TODO: generalize background material
    nn = np.arange(1, n + 1)
    k_d = np.sqrt(eps_r * mu_r) * k_0
    scale_term = (2 * nn + 1) / (nn * (nn + 1))

    den1 = np.sqrt(mu_r) * ric_h2n(nn, k_0 * a) * ric_jnp(nn, k_d * a) - np.sqrt(eps_r) * ric_h2np(
        nn, k_0 * a
    ) * ric_jn(nn, k_d * a)
    den2 = np.sqrt(eps_r) * ric_h2n(nn, k_0 * a) * ric_jnp(nn, k_d * a) - np.sqrt(mu_r) * ric_h2np(
        nn, k_0 * a
    ) * ric_jn(nn, k_d * a)

    an = (
        (1j**-nn)
        * scale_term
        * (
            np.sqrt(eps_r) * ric_jnp(nn, k_0 * a) * ric_jn(nn, k_d * a)
            - np.sqrt(mu_r) * ric_jn(nn, k_0 * a) * ric_jnp(nn, k_d * a)
        )
        / den1
    )
    bn = (
        (1j**-nn)
        * scale_term
        * (
            np.sqrt(mu_r) * ric_jnp(nn, k_0 * a) * ric_jn(nn, k_d * a)
            - np.sqrt(eps_r) * ric_jn(nn, k_0 * a) * ric_jnp(nn, k_d * a)
        )
        / den2
    )
    cn = (1j**-nn) * scale_term * 1j * np.sqrt(eps_r) * mu_r / den1
    dn = (1j**-nn) * scale_term * 1j * np.sqrt(eps_r) * mu_r / den2

    return an, bn, cn, dn


def _calculate_field(
    xyz: npt.NDArray[np.float64],
    k: float,
    eta: float,
    cn: npt.NDArray[np.complex128],
    dn: npt.NDArray[np.complex128],
    fn: Callable,
    dfn: Callable,
) -> tuple[npt.NDArray[np.complex128], npt.NDArray[np.complex128]]:
    r, theta, phi = scatsol.utils.cart2spherical(xyz).T
    nn = np.arange(1, cn.shape[0] + 1)
    theta_cos, phi_cos, phi_sin = np.cos(theta), np.cos(phi), np.sin(phi)

    rbesselterm = fn(nn[:, np.newaxis], k * r)
    rbesselpterm = dfn(nn[:, np.newaxis], k * r)

    p, _ = scatsol.utils.lpmn(1, cn.shape[0], theta_cos)

    p_theta_sin = np.zeros_like(p)
    p_theta_sin[0] = -1
    p_theta_sin[1] = -3 * theta_cos
    for n in range(2, p_theta_sin.shape[0] - 1):
        p_theta_sin[n] = (2 * n + 1) / n * theta_cos * p_theta_sin[n - 1] - (n + 1) / n * p_theta_sin[n - 2]

    dp_theta_sin = np.zeros_like(p)
    dp_theta_sin[0] = theta_cos
    for n in range(2, dp_theta_sin.shape[0]):
        dp_theta_sin[n - 1] = (n + 1) * p_theta_sin[n - 2] - n * np.cos(theta) * p_theta_sin[n - 1]
    dp_theta_sin = -dp_theta_sin

    e_field = np.empty((xyz.shape[0], 3), dtype=np.complex128)
    e_field[:, 0] = ((phi_cos) / (1j * (k * r) ** 2)) * (cn * nn * (nn + 1) @ (rbesselterm * p))
    e_field[:, 1] = -(phi_cos) / (k * r) * (1j * cn @ (rbesselpterm * dp_theta_sin) + dn @ (rbesselterm * p_theta_sin))
    e_field[:, 2] = +(phi_sin) / (k * r) * (1j * cn @ (rbesselpterm * p_theta_sin) + dn @ (rbesselterm * dp_theta_sin))

    h_field = np.empty((xyz.shape[0], 3), dtype=np.complex128)
    h_field[:, 0] = ((phi_sin) / (1j * (k * r) ** 2)) * (dn * nn * (nn + 1) @ (rbesselterm * p))
    h_field[:, 1] = -(phi_sin) / (k * r) * (1j * dn @ (rbesselpterm * dp_theta_sin) + cn @ (rbesselterm * p_theta_sin))
    h_field[:, 2] = -(phi_cos) / (k * r) * (1j * dn @ (rbesselpterm * p_theta_sin) + cn @ (rbesselterm * dp_theta_sin))
    h_field = h_field / eta

    return e_field, h_field



from typing import Callable
import numpy as np
import numpy.typing as npt
from scipy.special import h2vp, hankel2, jv, jvp

from scatsol.material import Material, Medium
import scatsol.utils


def mie_incident_field(
    xyz: npt.NDArray[np.float64],
    radius: float,
    frequency: float,
    background: Material,
    pol: str = "TM",
    *,
    n: int = 50,
):
    bg = Medium(background, frequency)
    kn = _an_inc(bg.k, radius, n)
    if pol == "TM":
        Ei, Hi = _calculate_field(xyz, bg.k, bg.eta, kn, jv, jvp)
    else:
        Hi, Ei = _calculate_field(xyz, bg.k, 1 / bg.eta, kn, jv, jvp)
        Ei = -Ei
    return Ei, Hi


def mie_total_field(
    xyz: npt.NDArray[np.float64],
    radius: float,
    frequency: float,
    background: Material,
    cylinder: Material | None = None,
    pol: str = "TM",
    *,
    n: int = 50,
) -> tuple[npt.NDArray[np.complex128], npt.NDArray[np.complex128]]:

    Erpz = np.zeros_like(xyz, dtype=np.complex128)
    Hrpz = np.zeros_like(xyz, dtype=np.complex128)
    mask = (xyz[:, :2] ** 2).sum(axis=1) > radius**2

    bg = Medium(background, frequency)
    if cylinder == None: # if PEC cylinder
        if pol == "TM":
            an = _an_cond_tm(bg.k, radius, n)
            Erpz[mask], Hrpz[mask] = _calculate_field(xyz[mask], bg.k, bg.eta, an, hankel2, h2vp)
        else:
            an = _an_cond_te(bg.k, radius, n)
            Hrpz[mask], Erpz[mask] = _calculate_field(xyz[mask], bg.k, 1 / bg.eta, an, hankel2, h2vp)
            Erpz[mask] = -Erpz[mask]
    else: # else diel cylinder
        c = Medium(cylinder, frequency)
        if pol == "TM":
            an, cn = _an_cn_diel_tm(bg.k, c.k, bg.eta, c.eta, radius, n)
            Erpz[mask], Hrpz[mask] = _calculate_field(xyz[mask], bg.k, bg.eta, an, hankel2, h2vp)
            Erpz[~mask], Hrpz[~mask] = _calculate_field(xyz[~mask], c.k, c.eta, cn, jv, jvp)
        else:
            an, cn = _an_cn_diel_tm(bg.k, c.k, 1 / bg.eta, 1 / c.eta, radius, n)
            Hrpz[mask], Erpz[mask] = _calculate_field(xyz[mask], bg.k, 1 / bg.eta, an, hankel2, h2vp)
            Hrpz[~mask], Erpz[~mask] = _calculate_field(xyz[~mask], c.k, 1 / c.eta, cn, jv, jvp)
            Erpz[mask] = -Erpz[mask]

    Ei, Hi = mie_incident_field(xyz[mask], radius, frequency, background, pol, n=n)
    Erpz[mask] += Ei
    Hrpz[mask] += Hi

    return Erpz, Hrpz


def _an_cond_tm(k: float, a: float, n: int) -> npt.NDArray[np.complex128]:
    nn = np.arange(0, n)
    return -(1j**-nn) * jv(nn, k * a) / hankel2(nn, k * a)


def _an_cn_diel_tm(
    k: float, kd: float, eta: float, etad: float, a: float, n: int
) -> tuple[npt.NDArray[np.complex128], npt.NDArray[np.complex128]]:
    nn = np.arange(0, n)
    den = eta * hankel2(nn, k * a) * jvp(nn, kd * a) - etad * h2vp(nn, k * a) * jv(nn, kd * a)
    num = eta * jv(nn, k * a) * jvp(nn, kd * a) - etad * jvp(nn, k * a) * jv(nn, kd * a)
    an = -(1j ** (-nn)) * num / den
    cn = (etad * 2 * (1j ** (-nn + 1))) / (np.pi * k * a * den)

    return an, cn


def _an_cond_te(k: float, a: float, n: int) -> npt.NDArray[np.complex128]:
    nn = np.arange(0, n)
    return -(1j**-nn) * jvp(nn, k * a) / h2vp(nn, k * a)


def _an_inc(k: float, a: float, n: int) -> npt.NDArray[np.complex128]:
    nn = np.arange(0, n)
    return 1j**-nn


def _calculate_field(
    xyz: npt.NDArray[np.float64], k: float, eta: float, an: npt.NDArray[np.complex128], fn: Callable, dfn: Callable
) -> tuple[npt.NDArray[np.complex128], npt.NDArray[np.complex128]]:
    nn = np.arange(0, an.shape[0])
    eps = 1.0 + np.heaviside(nn - 0.5, 1)
    rho, phi, _ = scatsol.utils.cart2cylindrical(xyz).T

    e_field = np.zeros_like(xyz, dtype=np.complex128)
    h_field = np.zeros_like(xyz, dtype=np.complex128)

    besselterm = fn(nn[:, None], k * rho)
    besselpterm = dfn(nn[:, None], k * rho)
    costerm = np.cos(nn[:, None] * phi)
    sinterm = np.sin(nn[:, None] * phi)
    e_field[:, 2] = an * eps @ (besselterm * costerm)
    h_field[:, 0] = (nn * an * eps @ (besselterm * sinterm)) / (1j * k * eta * rho)
    h_field[:, 1] = (k * an * eps @ (besselpterm * costerm)) / (1j * k * eta * rho)

    return e_field, h_field



from scatsol.constant import FREE_SPACE_PERMITTIVITY as EPSILON_0
from scatsol.constant import FREE_SPACE_PERMEABILITY as MU_0
import numpy as np


class Material:
    def __init__(self, epsilon_r: float = 1.0, mu_r: float = 1.0) -> None:
        self.epsilon_r = epsilon_r
        self.mu_r = mu_r

    @property
    def epsilon(self) -> float:
        return EPSILON_0 * self.epsilon_r

    @property
    def mu(self) -> float:
        return MU_0 * self.mu_r

    @property
    def eta(self) -> float:
        return np.sqrt(self.mu / self.epsilon)

    def __str__(self) -> str:
        return f"Material(epsilon_r={self.epsilon_r}, mu_r={self.mu_r})"

    def __rep__(self) -> str:
        return self.__str__()


class Medium:
    def __init__(self, material: Material, frequency: float = 300e6) -> None:
        self.material = material
        self.frequency = frequency

    @property
    def k(self) -> float:
        return (
            2.0
            * np.pi
            * self.frequency
            * np.sqrt(self.material.epsilon * self.material.mu)
        )

    @property
    def eta(self) -> float:
        return np.sqrt(self.material.mu / self.material.epsilon)

    @property
    def omega(self) -> float:
        return np.sqrt(self.material.mu / self.material.epsilon)



import numpy as np
import numpy.typing as npt
import scipy.special
from scipy.special import spherical_jn as sjn, spherical_yn as syn


def field_cyl2cart(field_rtz: npt.NDArray[np.complex128], xyz: npt.NDArray[np.float64]) -> npt.NDArray[np.complex128]:
    theta = np.arctan2(xyz[:, 1], xyz[:, 0])
    t_cos, t_sin = np.cos(theta), np.sin(theta)
    field_xyz = np.zeros_like(field_rtz)
    field_xyz[:, 0] = field_rtz[:, 0] * t_cos - field_rtz[:, 1] * t_sin
    field_xyz[:, 1] = field_rtz[:, 0] * t_sin + field_rtz[:, 1] * t_cos
    field_xyz[:, 2] = field_rtz[:, 2]
    return field_xyz


def field_sph2cart(field_rtz: npt.NDArray[np.complex128], xyz: npt.NDArray[np.float64]) -> npt.NDArray[np.complex128]:
    _, theta, phi = cart2spherical(xyz).T
    t_cos, t_sin = np.cos(theta), np.sin(theta)
    p_cos, p_sin = np.cos(phi), np.sin(phi)
    field_xyz = np.empty_like(field_rtz)
    field_xyz[:, 0] = field_rtz[:, 0] * t_sin * p_cos + field_rtz[:, 1] * t_cos * p_cos - field_rtz[:, 2] * p_sin
    field_xyz[:, 1] = field_rtz[:, 0] * t_sin * p_sin + field_rtz[:, 1] * t_cos * p_sin + field_rtz[:, 2] * p_cos
    field_xyz[:, 2] = field_rtz[:, 0] * t_cos - field_rtz[:, 1] * t_sin
    return field_xyz


def cart2spherical(xyz: npt.NDArray[np.float64]) -> npt.NDArray[np.float64]:
    rthetaphi = np.zeros(xyz.shape)
    xy = xyz[:, 0] ** 2 + xyz[:, 1] ** 2
    rthetaphi[:, 0] = np.sqrt(xy + xyz[:, 2] ** 2)
    rthetaphi[:, 1] = np.arctan2(np.sqrt(xy), xyz[:, 2])
    rthetaphi[:, 2] = np.arctan2(xyz[:, 1], xyz[:, 0])
    return rthetaphi


def cart2cylindrical(xyz: npt.NDArray[np.float64]) -> npt.NDArray[np.float64]:
    rphiz = np.zeros(xyz.shape)
    rphiz[:, 0] = np.sqrt(xyz[:, 0] ** 2 + xyz[:, 1] ** 2)
    rphiz[:, 1] = np.arctan2(xyz[:, 1], xyz[:, 0])
    rphiz[:, 2] = xyz[:, 2]
    return rphiz


def lpmn(m: int, n: int, x: np.ndarray) -> tuple[npt.NDArray[np.float64], ...]:
    p: npt.NDArray[np.float64] = np.zeros((n, x.shape[0]), dtype=np.float64)
    dp: npt.NDArray[np.float64] = np.zeros_like(p)
    for i in range(x.shape[0]):
        u, du = scipy.special.lpmn(m, n, x[i])
        p[:, i] = u[m, 1:]
        dp[:, i] = du[m, 1:]
    return p, dp


def ric_jn(n, z):
    return z * sjn(n, z)


def ric_jnp(n, z):
    return sjn(n, z) + z * sjn(n, z, derivative=True)


def ric_h2n(n, z):
    return z * (sjn(n, z) - 1j * syn(n, z))


def ric_h2np(n, z):
    return sjn(n, z) - 1j * syn(n, z) + z * (sjn(n, z, derivative=True) - 1j * syn(n, z, derivative=True))
